from typing import Optional, Any, Dict, List, Iterable, Union, Tuple

from pathlib import Path
import logging
from copy import deepcopy
from collections import OrderedDict

import networkx as nx
from networkx.drawing.nx_agraph import to_agraph

from canals.pipeline._utils import (
    PipelineRuntimeError,
    PipelineConnectError,
    PipelineValidationError,
    NoSuchStoreError,
    PipelineMaxLoops,
    locate_pipeline_input_nodes,
    locate_pipeline_output_nodes,
)


logger = logging.getLogger(__name__)


PYGRAPHVIZ_IMPORTED = False
try:
    import pygraphviz  # pylint: disable=unused-import

    PYGRAPHVIZ_IMPORTED = True
except ImportError:
    logger.info(
        "Could not import `pygraphviz`. Please install via: \n"
        "pip install pygraphviz\n"
        "(You might need to run this first: apt install libgraphviz-dev graphviz )"
    )


class Pipeline:
    """
    Nodes orchestration engine.

    Builds a graph of nodes and orchestrates their execution according to the execution graph.
    """

    def __init__(
        self,
        max_loops_allowed: int = 100,
    ):
        """
        Creates the Pipeline.

        :param max_loops_allowed: how many times the pipeline can run the same node before throwing an exception.
        """
        self.stores: Dict[str, object] = {}
        self.max_loops_allowed = max_loops_allowed
        self.graph = nx.DiGraph()

    def add_store(self, name: str, store: object) -> None:
        """
        Make a store available to all nodes of this pipeline.

        :param name: the name of the store.
        :param store: the store object.
        :returns: None
        """
        self.stores[name] = store

    def list_stores(self) -> Iterable[str]:
        """
        Returns a dictionary with all the stores that are attached to this Pipeline.

        :returns: a dictionary with all the stores attached to this Pipeline.
        """
        return self.stores.keys()

    def get_store(self, name: str) -> object:
        """
        Returns the store associated with the given name.

        :param name: the name of the store
        :returns: the store
        """
        try:
            return self.stores[name]
        except KeyError as e:
            raise NoSuchStoreError(f"No store named '{name}' is connected to this pipeline.") from e

    def add_node(
        self,
        name: str,
        instance: Any,
        parameters: Optional[Dict[str, Any]] = None,
        input_node: bool = False,
        output_node: bool = False,
    ) -> None:
        """
        Create a node for the given node. Nodes are not connected to anything by default:
        use `Pipeline.connect()` to connect nodes together.

        Node names must be unique, but node instances can be reused if needed.

        :param name: the name of the node.
        :param instance: the node instance.
        :param parameters: default parameters to pass to this node's instance only then this
            specific node is executed. These parameters are NOT shared across nodes that use
            the same instance.
        :param input_node: whether this node should receive the input data given to
            `Pipeline.run()` directly, regardless of its location in the Pipeline.
        :param output_node: whether the output of this node should be returned as output,
            regardless of its location in the Pipeline.
        :returns: None
        """
        # Node names are unique
        if name in self.graph.nodes:
            raise ValueError(f"Node named '{name}' already exists: choose another name.")

        # Node instances must be nodes
        if not hasattr(instance, "__canals_node__"):
            raise PipelineValidationError(
                f"'{type(instance)}' doesn't seem to be a node. Is this class decorated with @node?"
            )

        # Params must be a dict
        if parameters and not isinstance(parameters, dict):
            raise ValueError("'parameters' must be a dictionary.")

        # Add node to the graph, disconnected
        logger.debug("Adding node '%s' (%s)", name, instance)
        self.graph.add_node(
            name,
            instance=instance,
            visits=0,
            parameters=parameters,
            input_node=input_node,
            output_node=output_node,
        )

    def connect(self, connect_from: str, connect_to: str) -> None:
        """
        Connect nodes together. All nodes to connect must exist in the pipeline.
        If connecting to an node that has several output edges, specify its name with 'node_name.edge_name'.

        :param connect_from: the node that deliver the values. This can be either a single node name or
            can be in the format `node_name.edge_name` if the node has multiple outputs.
        :param connect_to: the node that receives the values. This is always just the node name.
        :returns: None
        """
        upstream_node_name = connect_from
        downstream_node_name = connect_to

        # Find out the name of the edge
        edge_name = None
        # Edges may be named explicitly by passing 'node_name.edge_name' to connect().
        # Specify the edge name for the upstream node only.
        if "." in upstream_node_name:
            upstream_node_name, edge_name = upstream_node_name.split(".", maxsplit=1)
            upstream_node = self.graph.nodes[upstream_node_name]["instance"]
        else:
            # If the edge had no explicit name and the upstream node has multiple outputs, raise an exception
            upstream_node = self.graph.nodes[upstream_node_name]["instance"]
            if len(upstream_node.outputs) != 1:
                raise PipelineConnectError(
                    f"Please specify which output of '{upstream_node_name}' "
                    f"'{downstream_node_name}' should connect to. Node '{upstream_node_name}' has the following "
                    f"outputs: {upstream_node.outputs}"
                )
            edge_name = upstream_node.outputs[0]

        # Remove edge name from downstream_node name (it's needed only when the node is upstream)
        downstream_node_name = downstream_node_name.split(".", maxsplit=2)[0]
        downstream_node = self.graph.nodes[downstream_node_name]["instance"]

        # All nodes names must be in the pipeline already
        if upstream_node_name not in self.graph.nodes:
            raise PipelineConnectError(f"'{upstream_node_name}' is not present in the pipeline.")
        if downstream_node_name not in self.graph.nodes:
            raise PipelineConnectError(f"'{downstream_node_name}' is not present in the pipeline.")

        # Check if the edge with that name already exists between those two nodes
        if any(
            edge[1] == downstream_node_name and edge[2]["label"] == edge_name
            for edge in self.graph.edges.data(nbunch=upstream_node_name)
        ):
            logger.info(
                "An edge called '%s' connecting node '%s' and node '%s' already exists: skipping.",
                edge_name,
                upstream_node_name,
                downstream_node_name,
            )
            return

        # Find all empty slots in the upstream and downstream nodes
        free_downstream_inputs = deepcopy(downstream_node.inputs)
        for _, __, data in self.graph.in_edges(downstream_node_name, data=True):
            position = free_downstream_inputs.index(data["label"])
            free_downstream_inputs.pop(position)

        free_upstream_outputs = deepcopy(upstream_node.outputs)
        for _, __, data in self.graph.out_edges(upstream_node_name, data=True):
            position = free_upstream_outputs.index(data["label"])
            free_upstream_outputs.pop(position)

        # Make sure the edge is connecting one free input to one free output
        if edge_name not in free_downstream_inputs or edge_name not in free_upstream_outputs:
            inputs_string = "\n".join(
                [
                    " - " + edge[2]["label"] + f" (taken by {edge[0]})"
                    for edge in self.graph.in_edges(downstream_node_name, data=True)
                ]
                + [f" - {free_in_edge} (free)" for free_in_edge in free_downstream_inputs]
            )
            outputs_string = "\n".join(
                [
                    " - " + edge[2]["label"] + f" (taken by {edge[1]})"
                    for edge in self.graph.out_edges(upstream_node_name, data=True)
                ]
                + [f" - {free_out_edge} (free)" for free_out_edge in free_upstream_outputs]
            )
            raise PipelineConnectError(
                f"Cannot connect '{upstream_node_name}' with '{downstream_node_name}' "
                f"with an edge named '{edge_name}': "
                f"their declared inputs and outputs do not match.\n"
                f"Upstream node '{upstream_node_name}' declared these outputs:\n{outputs_string}\n"
                f"Downstream node '{downstream_node_name}' declared these inputs:\n{inputs_string}\n"
            )
        # Create the edge
        logger.debug(
            "Connecting node '%s' to node '%s' along edge '%s'",
            upstream_node_name,
            downstream_node_name,
            edge_name,
        )
        self.graph.add_edge(upstream_node_name, downstream_node_name, label=edge_name)

    def get_node(self, name: str) -> Dict[str, Any]:
        """
        Returns all the data associated with a node.

        :param name: the name of the node
        :returns: a dictionary containing all data that was given to `add_node()`
        """
        candidates = [node for node in self.graph.nodes if node == name]
        if not candidates:
            raise ValueError(f"Node named {name} not found.")
        return self.graph.nodes[candidates[0]]

    def draw(self, path: Path) -> None:
        """
        Draws the pipeline. Requires `pygraphviz`.
        Run `pip install canals[draw]` to install missing dependencies.

        :param path: where to save the drawing.
        """
        if not PYGRAPHVIZ_IMPORTED:
            raise ImportError(
                "Could not import `pygraphviz`. Please install via: \n"
                "pip install pygraphviz\n"
                "(You might need to run this first: apt install libgraphviz-dev graphviz )"
            )
        graph = deepcopy(self.graph)

        input_nodes = locate_pipeline_input_nodes(graph)
        output_nodes = locate_pipeline_output_nodes(graph)

        # Draw the input
        graph.add_node("input", shape="plain")
        for node in input_nodes:
            for edge in graph.nodes[node]["instance"].inputs:
                graph.add_edge("input", node, label=edge)

        # Draw the output
        graph.add_node("output", shape="plain")
        for node in output_nodes:
            for edge in graph.nodes[node]["instance"].outputs:
                graph.add_edge(node, "output", label=edge)

        graphviz = to_agraph(graph)
        graphviz.layout("dot")
        graphviz.draw(path)
        logger.debug("Pipeline diagram saved at %s", path)

    def warm_up(self):
        """
        Make sure all nodes are warm.

        It's the node's responsibility to make sure this method can be called at every `Pipeline.run()`
        without re-initializing everything.
        """
        for node in self.graph.nodes:
            if hasattr(self.graph.nodes[node]["instance"], "warm_up"):
                self.graph.nodes[node]["instance"].warm_up()

    def run(
        self,
        data: Union[Dict[str, Any], List[Tuple[str, Any]]],
        parameters: Optional[Dict[str, Dict[str, Any]]] = None,
        debug: bool = False,
    ) -> Dict[str, Any]:
        """
        Runs the pipeline.

        :param data: the inputs to give to the input nodes of the Pipeline.
        :param parameters: a dictionary with all the parameters of all the nodes, namespaced by node.
        :param debug: whether to collect and return debug information.
        :returns: a dictionary with the outputs of the output nodes of the Pipeline.
        """
        parameters = self._validate_parameters(parameters=parameters)
        self.warm_up()
        self._validate_pipeline()

        # **** The Pipeline.run() algorithm ****
        #
        # Nodes are run as soon as an input for them appears in the inputs buffer.
        # When there's more than a node at once  in the buffer (which means some
        # branches are running in parallel or that there are loops) they are selected to
        # run in FIFO order by the `inputs_buffer` OrderedDict.
        #
        # Inputs are labeled with the name of the node they're aimed for:
        #
        #   ````
        #   inputs_buffer[target_node] = [(input_edge, input_value), ...]
        #   ```
        #
        # Nodes should wait until all the necessary input data has arrived before running.
        # If they're popped from the input_buffer before they're ready, they're put back in.
        # If the pipeline has branches of different lengths, it's possible that a node has to
        # wait a bit and let other nodes pass before receiving all the input data it needs.
        #
        # Data access:
        # - Name of the node       # self.graph.nodes  (List[str])
        # - Node instance          # self.graph.nodes[node]["instance"]
        # - Input nodes            # [e[0] for e in self.graph.in_edges(node)]
        # - Output nodes           # [e[1] for e in self.graph.out_edges(node)]
        # - Output edges           # [e[2]["label"] for e in self.graph.out_edges(node, data=True)]

        logger.info("Pipeline execution started.")
        inputs_buffer: OrderedDict = OrderedDict()

        for node_name in locate_pipeline_input_nodes(self.graph):
            # NOTE: We allow users to pass dictionaries just for convenience.
            # The real input format is List[Tuple[str, Any]], to allow several input edges to have the same name.
            if isinstance(data, dict):
                data = list(data.items())
            inputs_buffer[node_name] = {"data": data, "parameters": parameters}

        # *** PIPELINE EXECUTION LOOP ***
        # We select the nodes to run by checking which keys are set in the
        # inputs buffer. If the key exists, the node might be ready to run.
        pipeline_results: Dict[str, List[Dict[str, Any]]] = {}
        while inputs_buffer:
            logger.debug("> Current node queue: %s", inputs_buffer.keys())

            node_name, node_inputs = inputs_buffer.popitem(last=False)  # FIFO

            # Check if we looped over this node too many times
            self._check_max_loops(node_name)

            ready_to_run, inputs_buffer = self._ready_to_run(
                node_name=node_name, node_inputs=node_inputs, inputs_buffer=inputs_buffer
            )
            if not ready_to_run:
                continue

            # **** RUN THE NODE ****
            # It is our turn! The node is ready to run and all inputs are ready
            #
            # Let's raise the visits count
            self.graph.nodes[node_name]["visits"] += 1

            # Check for default parameters and add them to the parameter's dictionary
            # Default parameters are the one passed with the `pipeline.add_node()` method
            # and have lower priority with respect to parameters passed through `pipeline.run()`
            # or the modifications made by other nodes along the pipeline.
            if self.graph.nodes[node_name]["parameters"]:
                node_inputs["parameters"][node_name] = {
                    **(self.graph.nodes[node_name]["parameters"] or {}),
                    **parameters.get(node_name, {}),
                }

            # Get the node
            node_node = self.graph.nodes[node_name]["instance"]

            # Call the node
            try:
                logger.info(
                    "* Running %s (visits: %s)",
                    node_name,
                    self.graph.nodes[node_name]["visits"],
                )
                logger.debug("   '%s' inputs: %s", node_name, node_inputs)
                node_results: Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]
                node_results = node_node.run(
                    name=node_name,
                    data=node_inputs["data"],
                    parameters=node_inputs["parameters"],
                    stores=self.stores,
                )
                logger.debug("   '%s' outputs: %s\n", node_name, node_results)
            except Exception as e:
                raise PipelineRuntimeError(
                    f"{node_name} raised '{e.__class__.__name__}: {e}' \ninputs={node_inputs['data']}\nparameters={node_inputs.get('parameters', None)}\n\n"
                    "See the stacktrace above for more information."
                ) from e

            # **** PROCESS THE OUTPUT ****
            # The node run successfully. Let's store or distribute the output it produced, if it's valid.
            #
            # Type-check the output
            if (
                len(node_results) != 2
                or not isinstance(node_results[0], dict)
                or not isinstance(node_results[1], dict)
                or not all(isinstance(params, dict) for params in node_results[1])
            ):
                raise PipelineRuntimeError(
                    f"The node '{node_name}' did not return proper output. Check out the '@node' docstring."
                )

            # Process the output of the node
            if not self.graph.out_edges(node_name):
                # If there are no output edges, the output of this node is the output of the pipeline:
                # store it in pipeline_results.
                if not node_name in pipeline_results.keys():
                    pipeline_results[node_name] = []
                # We use append() to account for the case in which a node outputs several times
                # (for example, it can happen if there's a loop upstream). The list gets unwrapped before
                # returning it if there's only one output.
                pipeline_results[node_name].append(node_results[0])
            else:
                inputs_buffer = self._route_output(
                    node_results=node_results, node_name=node_name, inputs_buffer=inputs_buffer
                )

        logger.info("Pipeline executed successfully.")

        # Simplify output for single edge, single output pipelines
        pipeline_results = self._unwrap_results(pipeline_results)
        return pipeline_results

    def _validate_parameters(self, parameters: Optional[Dict[str, Dict[str, Any]]]) -> Dict[str, Dict[str, Any]]:
        """
        Validate the input parameters.
        """
        if not parameters:
            return {}

        if any(node not in self.graph.nodes for node in parameters.keys()):
            logging.warning(
                "You passed parameters for one or more node(s) that do not exist in the pipeline: %s",
                [node for node in parameters.keys() if node not in self.graph.nodes],
            )
        return parameters

    def _validate_pipeline(self):
        """
        Make sure the pipeline has at least one input node and one output node.
        """
        if not locate_pipeline_input_nodes(self.graph):
            raise ValueError("This pipeline has no input nodes!")

        if not locate_pipeline_output_nodes(self.graph):
            raise ValueError("This pipeline has no output nodes!")

    def _check_max_loops(self, node_name: str):
        """
        Verify whether this node run too many times.
        """
        if self.graph.nodes[node_name]["visits"] > self.max_loops_allowed:
            raise PipelineMaxLoops(f"Maximum loops count ({self.max_loops_allowed}) exceeded for node '{node_name}'.")

    def _ready_to_run(
        self, node_name: str, node_inputs: Dict[str, Any], inputs_buffer: OrderedDict
    ) -> Tuple[bool, OrderedDict]:
        """
        Verify whether a node is ready to run.

        Returns true if the node should run, false otherwise, and the updated inputs buffer.
        """
        # List all the inputs the current node should be waiting for.
        inputs_received = [i[0] for i in node_inputs["data"]]

        # We should be wait on all edges except for the downstream ones, to support loops.
        # This downstream check is enabled only for nodes taking more than one input
        # (the "entrance" of the loop).
        is_merge_node = len(self.graph.in_edges(node_name)) != 1
        data_to_wait_for = [
            (e[0], e[2]["label"])  # the node and the edge label
            for e in self.graph.in_edges(node_name, data=True)  # for all input edges
            # if there's a path in the graph leading back from the current node to the
            # input node, in case of multiple input nodes.
            if not is_merge_node or not nx.has_path(self.graph, node_name, e[0])
        ]

        if not data_to_wait_for:
            # This is an input node, so it's ready to run.
            logger.debug("'%s' is an input node and it's ready to run.")
            return (True, inputs_buffer)

        # Do we have all the inputs we expect?
        nodes_to_wait_for, inputs_to_wait_for = zip(*data_to_wait_for)
        if sorted(inputs_to_wait_for) == sorted(inputs_received):
            return (True, inputs_buffer)

        # This node is missing some inputs.
        #
        # Did all the upstream nodes run?
        if not all(self.graph.nodes[node_to_wait_for]["visits"] > 0 for node_to_wait_for in nodes_to_wait_for):
            # Some node upstream didn't run yet, so we should wait for them.
            logger.debug(
                "Putting '%s' back in the queue, some inputs are missing "
                "(inputs to wait for: %s, inputs_received: %s)",
                node_name,
                inputs_to_wait_for,
                inputs_received,
            )
            # Put back the node in the inputs buffer at the back...
            inputs_buffer[node_name] = node_inputs
            # ... and do not run this node (yet)
            return (False, inputs_buffer)

        # All upstream nodes run, so it **must** be our turn.
        #
        # Are we missing ALL inputs or just a few?
        if not inputs_received:
            # ALL upstream nodes have been skipped.
            #
            # Let's skip this node and add all downstream nodes to the queue.
            self.graph.nodes[node_name]["visits"] += 1
            logger.debug(
                "Skipping '%s', all input nodes were skipped and no inputs were received "
                "(skipped nodes: %s, inputs: %s)",
                node_name,
                nodes_to_wait_for,
                inputs_to_wait_for,
            )
            # Put all downstream nodes in the inputs buffer...
            downstream_nodes = [e[1] for e in self.graph.out_edges(node_name)]
            for downstream_node in downstream_nodes:
                if not downstream_node in inputs_buffer:
                    inputs_buffer[downstream_node] = {
                        "data": [],
                        "parameters": {},
                    }
            # ... and never run this node
            return (False, inputs_buffer)

        # If all nodes upstream have run and we received SOME input,
        # this is a merge node that was waiting on a node that has been skipped, so it's ready to run.
        # Let's pass None on the missing edges and go ahead.
        #
        # Example:
        #
        # --------------- value ----+
        #                           |
        #        +---X--- even ---+ |
        #        |                | |
        # -- parity_check         sum --
        #        |                 |
        #        +------ odd ------+
        #
        # If 'parity_check' produces output only on 'odd', 'sum' should run
        # with 'value' and 'odd' only, because 'even' will never arrive.
        #
        inputs_to_wait_for = list(inputs_to_wait_for)
        for input_expected in inputs_to_wait_for:
            if input_expected in inputs_received:
                inputs_to_wait_for.pop(inputs_to_wait_for.index(input_expected))
        logger.debug(
            "Some nodes upstream of '%s' were skipped, so some inputs will be None (missing inputs: %s)",
            node_name,
            inputs_to_wait_for,
        )
        for missing_input in inputs_to_wait_for:
            node_inputs["data"].append((missing_input, None))

        return (True, inputs_buffer)

    def _route_output(
        self,
        node_name: str,
        node_results: Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]],
        inputs_buffer: OrderedDict,
    ) -> OrderedDict:
        """
        Distrubute the outputs of the node into the input buffer of downstream nodes.

        Returns the updated inputs buffer.
        """
        # This is not a terminal node: find out where the output goes, to which nodes and along which edge
        is_decision_node_for_loop = (
            any(nx.has_path(self.graph, edge[1], node_name) for edge in self.graph.out_edges(node_name))
            and len(self.graph.out_edges(node_name)) > 1
        )
        for edge_data in self.graph.out_edges(node_name, data=True):
            edge = edge_data[2]["label"]
            target_node = edge_data[1]

            # If this is a decision node and a loop is involved, we add to the input buffer only the nodes
            # that received their expected output and we leave the others out of the queue.
            if is_decision_node_for_loop and not edge in node_results[0].keys():
                if nx.has_path(self.graph, target_node, node_name):
                    # In case we're choosing to leave a loop, do not put the loop's node in the buffer.
                    logger.debug(
                        "Not adding '%s' to the inputs buffer: we're leaving the loop.",
                        target_node,
                    )
                else:
                    # In case we're choosing to stay in a loop, do not put the external node in the buffer.
                    logger.debug(
                        "Not adding '%s' to the inputs buffer: we're staying in the loop.",
                        target_node,
                    )
            else:
                # In all other cases, populate the inputs buffer for all downstream nodes, setting None to any
                # edge that did not receive input.
                if not target_node in inputs_buffer:
                    inputs_buffer[target_node] = {
                        "data": []
                    }  # Create the buffer for the downstream node if it's not there yet
                if edge in node_results[0].keys():
                    inputs_buffer[target_node]["data"].append((edge, node_results[0][edge]))
                inputs_buffer[target_node]["parameters"] = node_results[1]

        return inputs_buffer

    def _unwrap_results(self, pipeline_results):
        if len(pipeline_results.keys()) == 1:
            pipeline_results = pipeline_results[list(pipeline_results.keys())[0]]  # type: ignore

            if len(pipeline_results) == 1:
                pipeline_results = pipeline_results[0]  # type: ignore

        return pipeline_results



import logging


logger = logging.getLogger(__name__)


class NodeError(Exception):
    pass


def node(class_):
    '''
    Marks a class as a node. Any class decorated with `@node`
    can be picked up by a Pipeline, serialized, deserialized, etc.

    All nodes MUST follow the contract below.
    This docstring is the source of truth for nodes contract.

    ```python
    node
    class MyNode:

        def __init__(self, model_name: str: "deepset-ai/a-model-name"):
            """
            Nodes should have an `__init__` method where they define:

            - `self.inputs = [<expected_input_edge_name(s)>]`:
                A list with all the edges they can possibly receive input from

            - `self.outputs = [<expected_output_edge_name(s)>]`:
                A list with the edges they might possibly produce as output

            - `self.init_parameters = {<init parameters>}`:
                Any state they wish to be persisted in their YAML serialization.
                These values will be given to the `__init__` method of a new instance
                when the pipeline is deserialized.

            The `__init__` must be extrememly lightweight, because it's a frequent
            operation during the construction and validation of the pipeline. If a node
            has some heavy state to initialize (models, backends, etc...) refer to the
            `warm_up()` method.
            """
            # Lightweight state can be initialized here, for example storing the model name
            # to be loaded later. See self.warm_up()
            self.model = None
            self.model_name = model_name
            self.how_many_times_have_I_been_called = 0

            # Contract - all three are mandatory.
            self.init_parameters = {"model_name": model_name}
            self.inputs = ["expected_input_edge_name"]
            self.outputs = ["expected_output_edge_name"]

        def warm_up(self):
            """
            Optional method.

            This method is called by Pipeline before the graph execution.
            Make sure to avoid double-initializations, because Pipeline will not keep
            track of which nodes it called `warm_up` on.
            """
            if not self.model:
                self.model = AutoModel.load_from_pretrained(self.model_name)

        def run(
            self,
            name: str,
            data: List[Tuple[str, Any]],
            parameters: Dict[str, Any],
            stores: Dict[str, Any],
        ):
            """
            Mandatory method.

            This is the method where the main functionality of the node should be carried out.
            It's called by `Pipeline.run()`, which passes the following parameters to it:

            - `name: str`: the name of the node. Allows the node to find its own parameters in the `parameters` dictionary (see below).

            - `data: List[Tuple[str, Any]]`: the input data.
                Pipeline guarantees that the following assert always passes: `assert self.inputs == [name for name, value in data]`,
                which means that:
                - `data` is of the same length as `self.inputs`.
                - `data` contains one tuple for each string stored in `self.inputs`.
                - no guarantee is given on the values of these tuples: notably, if there was a decision node upstream, some values might be `None`.
                For example, if a node declares `self.inputs = ["value", "value"]` (think of a Sum node), `data` might look like:
                - `[("value", 1), ("value", 10)]`
                - `[("value", None), ("value", 10)]`
                - `[("value", None), ("value", None)]`, or even
                - `[("value", 1), ("value", ["something", "unexpected"])]`
                but it will never look like:
                - `[("value", 1), ("value", 10), ("value", 100)]`,
                - `[("value": 15)]` or
                - `[("value": 15), ("unexpected", 10)]`.

            - `parameters: Dict[str, Dict[str, Any]]`: a dictionary of dictionaries with all the parameters for all nodes.
                Note that all nodes have access to all parameters for all other nodes: this might come handy to nodes like `Agent`s, that
                want to influence the behavior of nodes downstream.
                Nodes can access their own parameters using `name`, but they must not assume their name is present in the dictionary.
                Therefore the best way to get the parameters is with `my_parameters = parameters.get(name, {})`

            - `stores`: a dictionary of all the (Document)Stores connected to this pipeline.

            Pipeline expect the output of this function to be either a dictionary or a tuple.
            If it's a dictionary, it should always abide to the following format:

            `{output_name: output_value for output_name in <subset of self.expected_output>}`

            Which means that:
            - Nodes are not forced to produce output on all the expected outputs: for example nodes taking a decision, like classifiers,
                can produce output on a subset of the expected output edges and Pipeline will figure out the rest.
            - Nodes must not add any key in the data dictionary that is not present in `self.outputs`,

            Nodes may also want to return a tuple when they altered the content of `parameters` and want their changes to propagate
            downstream. In that case, the format is `(data, parameters)` where `data` follows the contract above and `parameters` should
            match the same format as it had in input, so `{"node_name": {"parameter_name": parameter_value, ...}, ...}`

            """
            self.how_many_times_have_I_been_called += 1

            value = data[0][1]
            print(f"Hello I'm {name}! This instance have been called {self.how_many_times_have_I_been_called} times and this is the value I received: {value}")

            return {"expected_output_edge_name": value}
    ```
    '''
    logger.debug("Registering %s as a node", class_)

    # '__canals_node__' is used to distinguish nodes from regular classes.
    # Its value is set to the desired node name: normally it is the class name, but it can technically be customized.
    class_.__canals_node__ = class_.__name__

    # Check for run()
    if not hasattr(class_, "run"):
        raise NodeError("Nodes must have a 'run()' method. See the docs for more information.")

    return class_



import logging
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.nodes import AddValue, Remainder, Double, Sum

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    add_one = AddValue(add=1, input="value")

    pipeline = Pipeline()
    pipeline.add_node("add_one", add_one)
    pipeline.add_node("remainder", Remainder(input="value", divisor=3))
    pipeline.add_node("add_ten", AddValue(add=10, input="0"))
    pipeline.add_node("double", Double(input="1", output="value"))
    pipeline.add_node("add_three", AddValue(add=3, input="2"))
    pipeline.add_node("add_one_again", add_one)
    pipeline.add_node("sum", Sum(inputs=["value"] * 4), input_node=True)

    pipeline.connect("add_one", "remainder")
    pipeline.connect("remainder.0", "add_ten")
    pipeline.connect("remainder.1", "double")
    pipeline.connect("add_ten", "sum")
    pipeline.connect("double", "sum")
    pipeline.connect("remainder.2", "add_three")
    pipeline.connect("add_three", "add_one_again")
    pipeline.connect("add_one_again", "sum")

    try:
        pipeline.draw(tmp_path / "decision_and_merge_pipeline.png")
    except ImportError:
        logging.warning("pygraphviz not found, pipeline is not being drawn.")

    results = pipeline.run({"value": 1})
    pprint(results)

    assert results == {"sum": 7}


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.nodes import AddValue, Double

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    pipeline = Pipeline()
    pipeline.add_node("first_addition", AddValue(add=2))
    pipeline.add_node("second_addition", AddValue(add=1))
    pipeline.add_node("double", Double(input="value"))
    pipeline.connect("first_addition", "double")
    pipeline.connect("double", "second_addition")

    try:
        pipeline.draw(tmp_path / "linear_pipeline.png")
    except ImportError:
        logging.warning("pygraphviz not found, pipeline is not being drawn.")

    results = pipeline.run({"value": 1})
    pprint(results)

    assert results == {"value": 7}


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



# SPDX-FileCopyrightText: 2023-present U.N. Owen <void@some.where>
#
# SPDX-License-Identifier: MIT
from canals.pipeline import Pipeline
from canals.node import node
from canals.store import StoreInMemory



from typing import Dict, Any, List, Tuple, Set

from canals import node


@node
class Repeat:
    """
    Repeats the input value on all outputs.

    Single input, multi output node. Order of outputs is irrelevant.
    Doesn't accept parameters. Doesn't use stores.

    :param input: the name of the input edge.
    :param outputs: the list of the output edges.
    """

    def __init__(self, input: str = "value", outputs: Set[str] = {"first", "second"}):
        self.init_parameters = {
            "input": input,
            "outputs": outputs,
        }
        self.inputs = [input]
        self.outputs = outputs

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        return ({output: data[0][1] for output in self.outputs}, parameters)


def test_repeat_default():
    node = Repeat()
    results = node.run(name="test_node", data=[("value", 10)], parameters={}, stores={})
    assert results == ({"first": 10, "second": 10}, {})
    assert node.init_parameters == {"input": "value", "outputs": {"first", "second"}}


def test_repeat_init_params():
    node = Repeat(input="test", outputs={"one", "two"})
    results = node.run(name="test_node", data=[("test", 10)], parameters={}, stores={})
    assert results == ({"one": 10, "two": 10}, {})
    assert node.init_parameters == {"input": "test", "outputs": {"one", "two"}}



from typing import Dict, Any, List, Tuple, Union, Callable, Optional
import sys
import builtins
from importlib import import_module

from canals import node


@node
class Accumulate:
    """
    Accumulates the value flowing through the edge into an internal attribute.

    Example of how to deal with serialization when some of the parameters
    are not directly serializable.

    Stateful, single input, single output node. Does not use stores.

    :param edge: the edge to read the value from.
    :param function: the function to use to accumulate the values.
        The function must take exactly two values.
        If it's a callable, it's used as it is.
        If it's a string, the node will look for it in sys.modules and
        import it at need. This is also a parameter.
    """

    def __init__(self, edge: str, function: Optional[Union[Callable, str]] = None):
        self.state = 0

        if function is None:
            self.function = lambda x, y: x + y
        else:
            self.function = self._load_function(function)

        self.init_parameters = {"edge": edge, "function": self._save_function(function) if function else None}
        self.inputs = [edge]
        self.outputs = [edge]

    def _load_function(self, function: Union[Callable, str]):
        """
        Loads the function by trying to import it.
        """
        if not isinstance(function, str):
            return function

        parts = function.split(".")
        module_name = ".".join(parts[:-1])
        function_name = parts[-1]

        module = import_module(module_name)
        return getattr(module, function_name)

    def _save_function(self, function: Union[Callable, str]):
        """
        Saves the function by returning its import path to be used with `_load_function`
        (which uses `import_module` internally).
        """
        if isinstance(function, str):
            return function
        module = sys.modules.get(function.__module__)
        if not module:
            raise ValueError("Could not locate the import module.")
        if module == builtins:
            return function.__name__
        return f"{module.__name__}.{function.__name__}"

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        function = parameters.get(name, {}).get("function", self.function)
        self.state = function(self.state, data[0][1])
        return ({data[0][0]: data[0][1]}, parameters)


def test_accumulate_default():
    node = Accumulate(edge="test")
    results = node.run(name="test_node", data=[("test", 10)], parameters={}, stores={})
    assert results == ({"test": 10}, {})
    assert node.state == 10
    assert node.init_parameters == {"edge": "test", "function": None}


def my_subtract(first, second):
    return first - second


def test_accumulate_callable():
    node = Accumulate(edge="test", function=my_subtract)
    results = node.run(name="test_node", data=[("test", 10)], parameters={}, stores={})
    assert results == ({"test": 10}, {})
    assert node.state == -10
    assert node.init_parameters == {"edge": "test", "function": "test.nodes.test_accumulate.my_subtract"}


def test_accumulate_string():
    node = Accumulate(edge="test", function="test.nodes.test_accumulate.my_subtract")
    results = node.run(name="test_node", data=[("test", 10)], parameters={}, stores={})
    assert results == ({"test": 10}, {})
    assert node.state == -10
    assert node.init_parameters == {"edge": "test", "function": "test.nodes.test_accumulate.my_subtract"}



from typing import Dict, Any, List, Tuple

from canals import node


@node
class AddValue:
    def __init__(self, add: int = 1, input: str = "value", output: str = "value"):
        """
        Adds the value of `add` to the value of the incoming edge.

        Single input, single output node. Doesn't use stores.

        :param add: the value to add. This is also a parameter.
        :param input: name of the input edge
        :param output: name of the output edge
        """
        self.add = add

        self.init_parameters = {"add": add, "input": input, "output": output}
        self.inputs = [input]
        self.outputs = [output]

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        sum = parameters.get(name, {}).get("add", self.add)
        sum += data[0][1]
        return ({self.outputs[0]: sum}, parameters)


def test_addvalue_default():
    node = AddValue()
    results = node.run(name="test_node", data=[("value", 10)], parameters={}, stores={})
    assert results == ({"value": 11}, {})
    assert node.init_parameters == {"add": 1, "input": "value", "output": "value"}


def test_addvalue_init_params():
    node = AddValue(add=3, input="test_in", output="test_out")
    results = node.run(name="test_node", data=[("test_in", 10)], parameters={}, stores={})
    assert results == ({"test_out": 13}, {})
    assert node.init_parameters == {"add": 3, "input": "test_in", "output": "test_out"}


def test_addvalue_runtime_params():
    node = AddValue(add=3, input="test_in", output="test_out")
    results = node.run(name="test_node", data=[("test_in", 10)], parameters={"test_node": {"add": 5}}, stores={})
    assert results == ({"test_out": 15}, {"test_node": {"add": 5}})
    assert node.init_parameters == {"add": 3, "input": "test_in", "output": "test_out"}



from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.nodes import AddValue, Sum

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):

    add_two = AddValue(add=2)
    make_the_sum = Sum(inputs=["value"] * 2)

    pipeline = Pipeline()
    pipeline.add_node("first_addition", add_two)
    pipeline.add_node("second_addition", add_two)
    pipeline.add_node("third_addition", add_two)
    pipeline.add_node("sum", make_the_sum)
    pipeline.add_node("fourth_addition", AddValue(add=1, input="sum"))

    pipeline.connect("first_addition", "second_addition")
    pipeline.connect("second_addition", "sum")
    pipeline.connect("third_addition", "sum")
    pipeline.connect("sum", "fourth_addition")

    try:
        pipeline.draw(tmp_path / "merging_pipeline.png")
    except ImportError:
        logging.warning("pygraphviz not found, pipeline is not being drawn.")

    results = pipeline.run({"value": 1})
    pprint(results)

    assert results == {"value": 9}


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



from typing import Dict, Any, List, Tuple

from canals import node


@node
class Merge:
    def __init__(
        self,
        inputs: List[str] = ["value"],
        output: str = "value",
    ):
        """
        Takes several input nodes and returns the first one that is not None.
        If no input edges received any value, returns None as well.

        Multi input, single output node. Doesn't take any parameter. Doesn't use stores.

        :param inputs: the inputs to expect.
        :param output: the name of the output edge.
        """
        self.output = output
        self.init_parameters = {
            "inputs": inputs,
            "output": output,
        }
        self.inputs = inputs
        self.outputs = [output]

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        for _, value in data:
            if value is not None:
                return ({self.outputs[0]: value}, parameters)


def test_merge_default():
    node = Merge()
    results = node.run(name="test_node", data=[("value", 5)], parameters={}, stores={})
    assert results == ({"value": 5}, {})
    assert node.init_parameters == {
        "inputs": ["value"],
        "output": "value",
    }


def test_merge_init_parameters():
    node = Merge(inputs=["test1", "test2", "test3"], output="test")
    results = node.run(name="test_node", data=[("test1", None), ("test2", 5), ("test3", 10)], parameters={}, stores={})
    assert results == ({"test": 5}, {})



from pathlib import Path
from pprint import pprint
import logging

from canals.pipeline import Pipeline
from test.nodes import Accumulate, AddValue, Greet, Remainder, Rename, Merge, Below, Double, Sum, Repeat, Subtract

logging.basicConfig(level=logging.DEBUG)


def test_complex_pipeline(tmp_path):
    accumulate = Accumulate(edge="value")

    pipeline = Pipeline(max_loops_allowed=4)
    pipeline.add_node("greet_first", Greet(edge="value", message="Hello!"))
    pipeline.add_node("accumulate_1", accumulate)
    pipeline.add_node("add_two", AddValue(add=2))
    pipeline.add_node("parity_check", Remainder(divisor=2))
    pipeline.add_node("add_one", AddValue(add=1, input="1"))
    pipeline.add_node("accumulate_2", accumulate)
    pipeline.add_node("rename_above_to_value", Rename(input="above", output="value"))

    pipeline.add_node("loop_merger", Merge(inputs=["value", "value"]))
    pipeline.add_node("below_10", Below(threshold=10))
    pipeline.add_node("double", Double(input="below", output="value"))

    pipeline.add_node("greet_again", Greet(edge="0", message="Hello again!"))
    pipeline.add_node("sum", Sum(inputs=["1", "0", "value"]))

    pipeline.add_node("greet_enumerator", Greet(edge="value", message="Hello from enumerator!"))
    pipeline.add_node("enumerate", Repeat(input="value", outputs=["0", "1"]))
    pipeline.add_node("add_three", AddValue(add=3, input="0"))

    pipeline.add_node("diff", Subtract(first_input="value", second_input="sum"))
    pipeline.add_node("greet_one_last_time", Greet(edge="diff", message="Bye bye!"))
    pipeline.add_node("replicate", Repeat(input="diff", outputs=["first", "second"]))
    pipeline.add_node("add_five", AddValue(add=5, input="first"))
    pipeline.add_node("add_four", AddValue(add=4, input="second"))
    pipeline.add_node("accumulate_3", accumulate)

    pipeline.connect("greet_first", "accumulate_1")
    pipeline.connect("accumulate_1", "add_two")
    pipeline.connect("add_two", "parity_check")

    pipeline.connect("parity_check.0", "greet_again")
    pipeline.connect("greet_again", "sum")
    pipeline.connect("sum", "diff")
    pipeline.connect("diff", "greet_one_last_time")
    pipeline.connect("greet_one_last_time", "replicate")
    pipeline.connect("replicate.first", "add_five")
    pipeline.connect("replicate.second", "add_four")
    pipeline.connect("add_four", "accumulate_3")

    pipeline.connect("parity_check.1", "add_one")
    pipeline.connect("add_one", "loop_merger")
    pipeline.connect("loop_merger", "below_10")

    pipeline.connect("below_10.below", "double")
    pipeline.connect("double", "loop_merger")

    pipeline.connect("below_10.above", "rename_above_to_value")
    pipeline.connect("rename_above_to_value", "accumulate_2")
    pipeline.connect("accumulate_2", "diff")

    pipeline.connect("greet_enumerator", "enumerate")
    pipeline.connect("enumerate.1", "sum")

    pipeline.connect("enumerate.0", "add_three")
    pipeline.connect("add_three", "sum")

    try:
        pipeline.draw(tmp_path / "complex_pipeline.png")
    except ImportError:
        logging.warning("pygraphviz not found, pipeline is not being drawn.")

    results = pipeline.run({"value": 1})
    pprint(results)
    print("accumulated: ", accumulate.state)

    assert results == {"add_five": [{"value": 16}], "accumulate_3": [{"value": 15}]}
    assert accumulate.state == 32


if __name__ == "__main__":
    test_complex_pipeline(Path(__file__).parent)



import pytest
from canals.store import MemoryStore, MissingItemError


def test_store():
    store = MemoryStore()

    assert not store.has_item("0")
    with pytest.raises(MissingItemError):
        store.get_item("0")
    store.delete_items(["0"])
    with pytest.raises(MissingItemError):
        store.delete_items(["0"], fail_on_missing_item=True)

    assert store.count_items() == 0
    assert store.get_ids() == []
    assert store.get_items() == []

    with pytest.raises(ValueError, match="must have an 'id' field"):
        store.write_items([{"key": "value"}])

    dictionary = {"id": "0", "key": "value"}
    store.write_items([dictionary])

    assert store.has_item("0")
    assert store.get_item("0") == dictionary
    assert store.count_items() == 1
    assert store.get_ids() == ["0"]
    assert store.get_items() == [dictionary]

    store.delete_items(["0"])
    store.delete_items(["0"])
    with pytest.raises(MissingItemError):
        store.delete_items("0", fail_on_missing_item=True)

    assert not store.has_item("0")
    with pytest.raises(MissingItemError):
        store.get_item("0")
    assert store.count_items() == 0
    assert store.get_ids() == []
    assert store.get_items() == []



from typing import Dict, Any, List, Tuple

from canals import node


@node
class Double:
    def __init__(self, input: str = "value", output: str = "value"):
        """
        Doubles the value in input.

        Single input single output node. Doesn't take parameters. Doesn't use stores.

        :param input: the name of the input.
        :param output: the name of the output.
        """
        self.init_parameters = {"input": input, "output": output}
        self.inputs = [input]
        self.outputs = [output]

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        for _, value in data:
            value *= 2

        return ({self.outputs[0]: value}, parameters)


def test_double_default():
    node = Double()
    results = node.run(name="test_node", data=[("value", 10)], parameters={}, stores={})
    assert results == ({"value": 20}, {})
    assert node.init_parameters == {"input": "value", "output": "value"}


def test_double_init_params():
    node = Double(input="test_in", output="test_out")
    results = node.run(name="test_node", data=[("test_in", 10)], parameters={}, stores={})
    assert results == ({"test_out": 20}, {})
    assert node.init_parameters == {"input": "test_in", "output": "test_out"}



from typing import Dict, Any, List, Tuple

from canals import node


@node
class Rename:
    """
    Renames an edge without changing the value.

    Single input, single output node. Doesn't accept parameters.
    Doesn't use stores.

    :param input: the input edge name.
    :param output: the output edge name.
    """

    def __init__(self, input: str, output: str):
        # Contract
        self.init_parameters = {"input": input, "output": output}
        self.inputs = [input]
        self.outputs = [output]

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        return ({self.outputs[0]: data[0][1]}, parameters)


def test_rename():
    node = Rename(input="test_in", output="test_out")
    results = node.run(name="test_node", data=[("test_in", 10)], parameters={}, stores={})
    assert results == ({"test_out": 10}, {})
    assert node.init_parameters == {"input": "test_in", "output": "test_out"}



from typing import Dict, Any, List, Tuple, Optional

import pytest

from canals import node


@node
class Remainder:
    """
    Redirects the value, unchanged, along the edge corresponding to the remainder
    of a division. For example, if `divisor=3`, the value `5` would be sent along
    the second output edge.

    Single input, multi output decision node. Order of output edges is critical.
    Doesn't accept parameters (divisor is tied to the number of output edges).
    Doesn't use stores.

    :param divisor: the number to divide the input value for.
    :param input: the name of the input edge.
    :param outputs: the name of the output edges. Must be equal in length to the
        divisor (if dividing by 3, you must give exactly three output names).
        Ordering is important.
    """

    def __init__(self, divisor: int = 2, input: str = "value", outputs: Optional[List[str]] = None):
        if outputs and len(outputs) != divisor:
            raise ValueError("The divisor must be equal to the number of output edges.")
        if not outputs:
            outputs = [str(remainder) for remainder in range(divisor)]
        self.divisor = divisor

        self.init_parameters = {"divisor": divisor, "input": input, "outputs": outputs}
        self.inputs = [input]
        self.outputs = outputs

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        remainder = data[0][1] % self.divisor
        return ({self.outputs[remainder]: data[0][1]}, parameters)


def test_remainder_default():
    node = Remainder()
    results = node.run(name="test_node", data=[("value", 10)], parameters={}, stores={})
    assert results == ({"0": 10}, {})

    results = node.run(name="test_node", data=[("value", 11)], parameters={}, stores={})
    assert results == ({"1": 11}, {})
    assert node.init_parameters == {"divisor": 2, "input": "value", "outputs": ["0", "1"]}


def test_remainder_default_output_for_divisor():
    node = Remainder(divisor=5)
    results = node.run(name="test_node", data=[("value", 10)], parameters={}, stores={})
    assert results == ({"0": 10}, {})

    results = node.run(name="test_node", data=[("value", 13)], parameters={}, stores={})
    assert results == ({"3": 13}, {})
    assert node.init_parameters == {"divisor": 5, "input": "value", "outputs": ["0", "1", "2", "3", "4"]}


def test_remainder_init_params():
    with pytest.raises(ValueError):
        node = Remainder(divisor=3, input="test", outputs=["one", "two"])

    with pytest.raises(ValueError):
        node = Remainder(divisor=3, input="test", outputs=["zero", "one", "two", "three"])

    node = Remainder(divisor=3, input="test", outputs=["zero", "one", "two"])
    results = node.run(name="test_node", data=[("value", 10)], parameters={}, stores={})
    assert results == ({"one": 10}, {})
    assert node.init_parameters == {"divisor": 3, "input": "test", "outputs": ["zero", "one", "two"]}



from typing import Dict, Any, List, Tuple

from canals import node


@node
class Sum:
    def __init__(self, inputs: List[str] = ["value"], output: str = "sum"):
        """
        Sums the values of all the input edges together.

        Multi input, single output node. Order of input edges is irrelevant.
        Doesn't use stores. Doesn't have parameters.

        :param inputs: list of edges to sum.
        :param output: name of the output edge.
        """
        self.init_parameters = {"inputs": inputs, "output": output}
        self.inputs = inputs
        self.outputs = [output]

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        sum = 0
        for _, value in data:
            if value:
                sum += value

        return ({self.outputs[0]: sum}, parameters)


def test_sum_default():
    node = Sum()
    results = node.run(name="test_node", data=[("value", 10)], parameters={}, stores={})
    assert results == ({"sum": 10}, {})
    assert node.init_parameters == {"inputs": ["value"], "output": "sum"}


def test_sum_init_params():
    node = Sum(inputs=["value", "value"], output="test_out")
    results = node.run(name="test_node", data=[("value", 10), ("value", 4)], parameters={}, stores={})
    assert results == ({"test_out": 14}, {})
    assert node.init_parameters == {"inputs": ["value", "value"], "output": "test_out"}



from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.nodes import AddValue, Repeat, Double

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    add_one = AddValue(add=1, input="value")

    pipeline = Pipeline()
    pipeline.add_node("add_one", add_one)
    pipeline.add_node("enumerate", Repeat(input="value", outputs=["0", "1", "2"]))
    pipeline.add_node("add_ten", AddValue(add=10, input="0"))
    pipeline.add_node("double", Double(input="1", output="value"))
    pipeline.add_node("add_three", AddValue(add=3, input="2"))
    pipeline.add_node("add_one_again", add_one)

    pipeline.connect("add_one", "enumerate.1")
    pipeline.connect("enumerate.0", "add_ten")
    pipeline.connect("enumerate.1", "double")
    pipeline.connect("enumerate.2", "add_three")
    pipeline.connect("add_three", "add_one_again")

    try:
        pipeline.draw(tmp_path / "parallel_branches_pipeline.png")
    except ImportError:
        logging.warning("pygraphviz not found, pipeline is not being drawn.")

    results = pipeline.run({"value": 1})
    pprint(results)

    assert results == {
        "add_one_again": [{"value": 6}],
        "add_ten": [{"value": 12}],
        "double": [{"value": 4}],
    }


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



from typing import Dict, Any, List, Tuple

from canals import node


@node
class Subtract:
    def __init__(self, first_input: str, second_input: str, output: str = "diff"):
        """
        Subtracts the second edge's value from the first.

        Double input, single output node. Order of input edges is critical.
        Doesn't have parameters. Doesn't use stores.

        :param first_input: name of the edge carrying the value to subtract from.
        :param second_input: name of the edge carrying the value to subtract.
        :param output: name of the output edge.
        """
        self.init_parameters = {"first_input": first_input, "second_input": second_input, "output": output}
        self.inputs = [first_input, second_input]
        self.outputs = [output]

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        first_value = [value for name, value in data if name == self.inputs[0]][0]
        second_value = [value for name, value in data if name == self.inputs[1]][0]

        return ({"diff": first_value - second_value}, parameters)


def test_subtract():
    node = Subtract(first_input="first", second_input="second", output="diff")
    results = node.run(name="test_node", data=[("second", 7), ("first", 10)], parameters={}, stores={})
    assert results == ({"diff": 3}, {})
    assert node.init_parameters == {"first_input": "first", "second_input": "second", "output": "diff"}



from typing import Dict, Any, List, Tuple
import logging

from canals import node


logger = logging.getLogger(__name__)


@node
class Greet:
    def __init__(self, message: str = "\nGreeting node says: Hi!\n", edge: str = "value"):
        """
        Logs a greeting message without affecting the value
        passing on the edge.

        Single input, single output node. Doesn't use stores.

        :param message: the message to print. This is also a parameter.
        :param edge: the name of the input and output edge.
        """
        self.message = message

        self.init_parameters = {"edge": edge, "message": message}
        self.inputs = [edge]
        self.outputs = [edge]

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        message = parameters.get(name, {}).get("message", self.message)
        logger.info(message)
        return ({data[0][0]: data[0][1]}, parameters)


def test_greet_default(caplog):
    caplog.set_level(logging.INFO)
    node = Greet()
    results = node.run(name="test_node", data=[("value", 10)], parameters={}, stores={})
    assert results == ({"value": 10}, {})
    assert "Greeting node says: Hi!" in caplog.text
    assert node.init_parameters == {"edge": "value", "message": "\nGreeting node says: Hi!\n"}


def test_greet_init_params(caplog):
    caplog.set_level(logging.INFO)
    node = Greet(edge="test", message="Hello!")
    results = node.run(name="test_node", data=[("test", 10)], parameters={}, stores={})
    assert results == ({"test": 10}, {})
    assert "Hello" in caplog.text
    assert node.init_parameters == {"edge": "test", "message": "Hello!"}


def test_greet_runtime_params(caplog):
    caplog.set_level(logging.INFO)
    node = Greet(edge="test", message="Hello!")
    results = node.run(
        name="test_node", data=[("test", 10)], parameters={"test_node": {"message": "Ciao :)"}}, stores={}
    )
    assert results == ({"test": 10}, {"test_node": {"message": "Ciao :)"}})
    assert "Ciao :)" in caplog.text
    assert node.init_parameters == {"edge": "test", "message": "Hello!"}



from typing import *
from pathlib import Path
from pprint import pprint

from canals.pipeline import Pipeline
from test.nodes import Accumulate, AddValue, Below, Merge, Sum

import logging

logging.basicConfig(level=logging.DEBUG)


def test_pipeline(tmp_path):
    accumulator = Accumulate(edge="value")

    pipeline = Pipeline(max_loops_allowed=10)
    pipeline.add_node("merge", Merge(), input_node=True)
    pipeline.add_node("sum", Sum(inputs=["value", "value"]), input_node=True)
    pipeline.add_node("below_10", Below(threshold=10))
    pipeline.add_node("add_one", AddValue(add=1, input="below"))
    pipeline.add_node("counter", accumulator)
    pipeline.add_node("add_two", AddValue(add=2, input="above"))

    pipeline.connect("merge", "below_10")
    pipeline.connect("below_10.below", "add_one")
    pipeline.connect("add_one", "counter")
    pipeline.connect("counter", "merge")
    pipeline.connect("below_10.above", "add_two")
    pipeline.connect("add_two", "sum")

    pipeline.draw(tmp_path / "looping_and_merge_pipeline.png")

    results = pipeline.run(
        {"value": 3},
    )
    pprint(results)
    print("accumulate: ", accumulator.state)

    assert results == {"sum": 15}
    assert accumulator.state == 7


if __name__ == "__main__":
    test_pipeline(Path(__file__).parent)



from typing import Literal, Any, Dict, List, Optional

import logging


logger = logging.getLogger(__name__)


class StoreError(Exception):
    pass


class DuplicateError(StoreError):
    pass


class MissingItemError(StoreError):
    pass


DuplicatePolicy = Literal["skip", "overwrite", "fail"]


class StoreInMemory:
    """
    Stores data in-memory. It's ephemeral and cannot be saved to disk.

    Note: this is a toy implementation meant to showcase the contract.
    Can still be used for small applications and demos.
    """

    def __init__(self):
        """
        Initializes the store.
        """
        self.storage = {}

    def has_item(self, object_id: str) -> bool:
        """
        Checks if this ID exists in the store.

        :param object_id: the object_id to find in the store.
        """
        return object_id in self.storage.keys()

    def get_item(self, object_id: str) -> Dict[str, Any]:
        """
        Finds a item by ID in the store. Fails if the item is not present.

        :param object_id: the object_id of the item to get.
        """
        if not self.has_item(object_id=object_id):
            raise MissingItemError(f"ID {object_id} not found.")
        return self.storage[object_id]

    def count_items(self, filters: Optional[Dict[str, Any]] = None) -> int:
        """
        Returns the number of how many items match the given filters.
        Pass filters={} to count all items in the store.

        :param filters: the filters to apply to the items list.
        """
        # TODO apply filters
        return len(self.storage.keys())

    def get_ids(self, filters: Optional[Dict[str, Any]] = None) -> List[str]:
        """
        Returns only the IDs of the items that match the filters provobject_ided.

        :param filters: the filters to apply to the item list.
        """
        # TODO apply filters
        return list(self.storage.keys())

    def get_items(self, filters: Optional[Dict[str, Any]] = None) -> List[Dict[str, Any]]:
        """
        Returns the items that match the filters provobject_ided.

        :param filters: the filters to apply to the item list.
        """
        # TODO apply filters
        return [self.storage[object_id] for object_id in self.get_ids(filters=filters)]

    def write_items(self, items: List[Dict[str, Any]], duplicates: DuplicatePolicy = "fail") -> None:
        """
        Writes (or overwrites) items into the store.

        :param items: a list of dictionaries.
        :param duplicates: items with the same ID count as duplicates. When duplicates are met,
            the store can:
             - skip: keep the existing item and ignore the new one.
             - overwrite: remove the old item and write the new one.
             - fail: an error is raised
        :raises DuplicateError: Exception trigger on duplicate item
        :return: None
        """
        if not isinstance(items, list):
            raise ValueError("Please provobject_ide a list of dictionaries.")
        for item in items:
            if not "id" in item.keys():
                raise ValueError("Objects must have an 'id' field.")

            if self.has_item(item["id"]):
                if duplicates == "fail":
                    raise DuplicateError(f"ID {item['id']} already exists.")
                if duplicates == "skip":
                    logger.warning("ID '%s' already exists", item["id"])
            self.storage[item["id"]] = item

    def delete_items(self, object_ids: List[str], fail_on_missing_item: bool = False) -> None:
        """
        Deletes all object_ids from the given pool.

        :param object_ids: the object_ids to delete
        :param fail_on_missing_item: fail if the object_id is not found, log ignore otherwise
        """
        for object_id in object_ids:
            if not self.has_item(object_id=object_id):
                if fail_on_missing_item:
                    raise MissingItemError(f"ID {object_id} not found, cannot delete it.")
                logger.info("ID %s not found, cannot delete it.", object_id)
                return
            del self.storage[object_id]



from typing import Dict, TYPE_CHECKING

import json
import logging
from pathlib import Path

if TYPE_CHECKING:
    from canals import Pipeline


logger = logging.getLogger(__name__)


class PipelineError(Exception):
    pass


class NoSuchStoreError(PipelineError):
    pass


class PipelineRuntimeError(Exception):
    pass


class PipelineConnectError(PipelineError):
    pass


class PipelineValidationError(PipelineError):
    pass


class PipelineMaxLoops(PipelineError):
    pass


def locate_pipeline_input_nodes(graph):
    """
    Collect the nodes with no input edges: they receive directly the pipeline inputs.
    """
    return [node for node in graph.nodes if not graph.in_edges(node) or graph.nodes[node]["input_node"]]


def locate_pipeline_output_nodes(graph):
    """
    Collect the nodes with no output edges: these define the output of the pipeline.
    """
    return [node for node in graph.nodes if not graph.out_edges(node) or graph.nodes[node]["output_node"]]


def save_pipelines(pipelines: Dict[str, "Pipeline"], path: Path, _writer=json.dump) -> None:
    """
    Converts a dictionary of named Pipelines into a Python dictionary that can be
    written to a JSON file.

    :param pipelines: dictionary of {name: pipeline_object}
    :param path: where to write the resulting file
    :param _writer: which function to use to write the dictionary to a file.
        Use this parameter to dump to a different format like YAML, TOML, HCL, etc.
    """
    return


def load_pipelines(path: Path, _reader=json.load) -> Dict[str, "Pipeline"]:
    """
    Reads the given file and returns a dictionary of named Pipelines ready to use.

    :param path: the path of the pipelines file.
    :param _reader: which function to use to read the dictionary from file.
        Use this parameter to read from a different format like YAML, TOML, HCL, etc.
    """
    return {}



from test.nodes.test_accumulate import Accumulate
from test.nodes.test_add_value import AddValue
from test.nodes.test_below import Below
from test.nodes.test_double import Double
from test.nodes.test_greet import Greet
from test.nodes.test_merge import Merge
from test.nodes.test_remainder import Remainder
from test.nodes.test_rename import Rename
from test.nodes.test_repeat import Repeat
from test.nodes.test_subtract import Subtract
from test.nodes.test_sum import Sum



from typing import Dict, Any, List, Tuple

from canals import node


@node
class Below:
    def __init__(
        self,
        threshold: int,
        input: str = "value",
        output_above: str = "above",
        output_below: str = "below",
    ):
        """
        Redirects the value, unchanged, along a different edge whether the value is above
        or below the given threshold.

        Single input, double output decision node. Doesn't use stores.

        :param threshold: the number to compare the input value against. This is also a parameter.
        :param input: the name of the input edge.
        :param output_above: the name of the output edge if the value is above the threshold.
        :param output_below: the name of the output edge if the value is below the threshold.
        """
        self.threshold = threshold
        self.output_above = output_above
        self.output_below = output_below

        self.init_parameters = {
            "threshold": threshold,
            "input": input,
            "output_above": output_above,
            "output_below": output_below,
        }
        self.inputs = [input]
        self.outputs = [output_above, output_below]

    def run(
        self,
        name: str,
        data: List[Tuple[str, Any]],
        parameters: Dict[str, Any],
        stores: Dict[str, Any],
    ):
        threshold = parameters.get(name, {}).get("threshold", self.threshold)
        if data[0][1] < threshold:
            return ({self.output_below: data[0][1]}, parameters)
        else:
            return ({self.output_above: data[0][1]}, parameters)


def test_below_default():
    node = Below(threshold=10)
    results = node.run(name="test_node", data=[("value", 5)], parameters={}, stores={})
    assert results == ({"below": 5}, {})

    results = node.run(name="test_node", data=[("value", 15)], parameters={}, stores={})
    assert results == ({"above": 15}, {})

    assert node.init_parameters == {
        "threshold": 10,
        "input": "value",
        "output_above": "above",
        "output_below": "below",
    }


def test_below_init_parameters():
    node = Below(threshold=10, input="test", output_above="higher", output_below="lower")
    results = node.run(name="test_node", data=[("test", 5)], parameters={}, stores={})
    assert results == ({"lower": 5}, {})

    results = node.run(name="test_node", data=[("test", 15)], parameters={}, stores={})
    assert results == ({"higher": 15}, {})

    assert node.init_parameters == {
        "threshold": 10,
        "input": "test",
        "output_above": "higher",
        "output_below": "lower",
    }


def test_below_run_parameters():
    node = Below(threshold=10, input="test", output_above="higher", output_below="lower")
    results = node.run(name="test_node", data=[("test", 5)], parameters={"test_node": {"threshold": 1}}, stores={})
    assert results == ({"higher": 5}, {"test_node": {"threshold": 1}})

    results = node.run(name="test_node", data=[("test", 15)], parameters={"test_node": {"threshold": 100}}, stores={})
    assert results == ({"lower": 15}, {"test_node": {"threshold": 100}})

    assert node.init_parameters == {
        "threshold": 10,
        "input": "test",
        "output_above": "higher",
        "output_below": "lower",
    }
